function MediaManager (project) {
  this.project = project;
  this.ds = project.ds;
  this.rdf = getRDFService();

  this.checkMediaAttributes();
}

MediaManager.prototype = {
  /**
   * Set the width and height attributes for all images.
   */
  checkMediaAttributes: function () {
    var IRes = Components.interfaces.nsIRDFResource;
    var typearc = this.rdf.GetResource(Cx.NS_RDF + "type");
    var widtharc = this.rdf.GetResource(Cx.NS_CX + "width");
    var heightarc = this.rdf.GetResource(Cx.NS_CX + "height");
    var imagetype = this.rdf.GetResource(Cx.NS_CX + "Image");
    var images = this.ds.GetSources(typearc, imagetype, true);
    while (images.hasMoreElements()) {
      var imageres = images.getNext().QueryInterface(IRes);
      if (this.ds.hasArcOut(imageres, widtharc) &&
          this.ds.hasArcOut(imageres, heightarc))
        continue;

      var file = this.fileForMedia(imageres);
      var nameparts = file.leafName.match(/^(.+)\.([^.]+)$/);
      if (nameparts.length != 3)
        continue;

      var ext = nameparts[2].toLowerCase();
      if (ext == "jpg")
        ext = "jpeg";
      try {
        var innercodecstr = "@mozilla.org/image/decoder;2?type=image/" + ext;
        var innercodec = Components.classes[innercodecstr]
          .createInstance(Components.interfaces.imgIDecoder);
        // var codecstr = "@greyfirst.ca/imgscale/imagecodec;1?type=" + ext;
        var codecstr = "@greyfirst.ca/imgscale/imagecodec;1";
        var codec = Components.classes[codecstr]
          .createInstance(Components.interfaces.imgscaleIImageCodec);
        var self = this;
        var observer = {
          onDecodeSucceeded: function (image) {
            dump("--- setting image size to " + image.width + "x" + image.height + "\n");
            setRDFString(self.ds, imageres, widtharc, image.width);
            setRDFString(self.ds, imageres, heightarc, image.height);
          },
          onDecodeFailed: function () {
            dump("*** Failed to decode " + file.path + "\n");
          }
        };
        dump("--- decoding " + file.leafName + " to establish size\n");
        codec.init(innercodec);
        var image = codec.decode(getBufferedFileInputStream(file), observer);
        if (image)
          observer.onDecodeSucceeded(image);
      }
      catch (ex) {
        dump("*** checkMediaAttributes: " + ex + "\n");
      }
    }
  },


  uriForMedia: function (mediares) {
    return fileToFileURL(this.fileForMedia(mediares));
  },


  fileForMedia: function (mediares) {
    var filearc = this.rdf.GetResource(Cx.NS_CX + "localFile");
    var filename = getRDFString(this.ds, mediares, filearc);
    if (! filename)
      throw "No localFile for media";
    var file = this.project.projectFolder;
    file.append(filename);
    return file;
  },


  mediaForFilename: function (filename) {
    var filearc = this.rdf.GetResource(Cx.NS_CX + "localFile");
    var filelit = this.rdf.GetLiteral(filename);
    var sources = this.ds.GetSources(filearc, filelit, true);
    if (sources.hasMoreElements())
      return sources.getNext().QueryInterface(
        Components.interfaces.nsIRDFResource);
    // Try a sanitized version
    filelit = this.rdf.GetLiteral(safeFileName(filename));
    sources = this.ds.GetSources(filearc, filelit, true);
    if (sources.hasMoreElements())
      return sources.getNext().QueryInterface(
        Components.interfaces.nsIRDFResource);
    return null;
  },


  /**
   * Create a thumbnail for the given media resource, with optional
   * constraints on width or height. It only works for images currently.
   * @param mediares  the media resource
   * @param maxwidth  the maximum width of the thumbnail (optional)
   * @param maxheight  the maximum height of the thumbnail (optional)
   * @type nsIFile
   * @return the thumbnail file
   */
  createThumbnail: function (mediares, maxwidth, maxheight, callback) {
    if (! maxwidth)
      maxwidth = 256;
    if (! maxheight)
      maxheight = 256;

    var file = this.fileForMedia(mediares);
    var nameparts = file.leafName.match(/^(.+)\.([^.]+)$/);
    if (nameparts.length != 3) {
      dump("*** createThumbnail: " + file.leafName
        + " is not a recognizable image name\n");
      callback(false);
      return;
    }

    var thumb = file.clone();
    thumb.leafName = nameparts[1] + "_thumb.png";
    var ext = nameparts[2].toLowerCase();
    if (ext == "jpg")
      ext = "jpeg";

    var innercodecstr = "@mozilla.org/image/decoder;2?type=image/" + ext;
    var innercodec = Components.classes[innercodecstr]
      .createInstance(Components.interfaces.imgIDecoder);
    //var codecstr = "@greyfirst.ca/imgscale/imagecodec;1?type=" + ext;
    var codecstr = "@greyfirst.ca/imgscale/imagecodec;1";
    var codec = Components.classes[codecstr]
      .createInstance(Components.interfaces.imgscaleIImageCodec);
    var self = this;
    var observer = {
      onDecodeSucceeded: function (image) {
        self.createThumbnail_decoded(mediares, maxwidth, maxheight,
          image, file, thumb, callback);
      },
      onDecodeFailed: function () {
        dump("*** createThumbnail: decode failed!\n");
        callback(false);
      }
    };
    codec.init(innercodec);
    // Handle the sync/async variant (returns null if async)
    dump("--- creating thumbnail for " + file.leafName + "\n");
    var image = codec.decode(getBufferedFileInputStream(file), observer);
    if (image)
      this.createThumbnail_decoded(mediares, maxwidth, maxheight, image,
        file, thumb, callback);
  },


  createThumbnail_decoded: function (mediares, maxwidth, maxheight,
      image, file, thumb, callback) {
    var width = image.width;
    var height = image.height;
    if (height > maxheight && width > maxwidth) {
      var aspect = width / height;
      if (width > height) {
        width = maxwidth;
        height = Math.floor(maxwidth / aspect);
        if (height > maxheight) {
          width = Math.floor(height * aspect);
          height = maxheight;
        }
      }
      else {
        height = maxheight;
        width = Math.floor(maxheight * aspect);
        if (width > maxwidth) {
          height = Math.floor(width / aspect);
          width = maxwidth;
        }
      }
    }

    if (! thumb.exists())
      thumb.create(0, 0600);

    var self = this;
    var observer = {
      onScaleSucceeded: function () {
        // Consideration: Should we allow multiple thumbnails per image and
        // point at a resource with width and height attributes?
        var thumbarc = self.rdf.GetResource(Cx.NS_CX + "thumbnail");
        setRDFString(self.ds, mediares, thumbarc, thumb.leafName);
        callback(thumb);
      },
      onScaleFailed: function () {
        dump("*** thumbnail creation failed\n");
        callback(false);
      }
    };

    var scaler = Components.classes["@greyfirst.ca/imgscale/imagescaler;1"]
      .createInstance(Components.interfaces.imgscaleIImageScaler);
    scaler.init(scaler.bicubic, observer);
    scaler.scale(fileToFileURL(file), fileToFileURL(thumb), width, height);
  },


  getThumbnail: function (mediares, force, maxwidth, maxheight, callback) {
    var thumbarc = this.rdf.GetResource(Cx.NS_CX + "thumbnail");
    var thumbname = getRDFString(this.ds, mediares, thumbarc);
    if (! thumbname)
      return force ? this.createThumbnail(mediares, maxwidth, maxheight, callback) : null;
    var file = this.project.projectFolder;
    file.append(thumbname);
    if (! file.exists() && force)
      return this.createThumbnail(mediares, maxwidth, maxheight, callback);
    callback(file);
    return file;
  },


  /**
   * This function should only be used for creating a media resource
   * for media files that have already been added to the project, but
   * were not added with a corresponding media resource. This is not
   * the correct way to add new media files.
   * @see #addMediaFromFile
   * @param filename  the name of the file in the project folder
   * @type nsIRDFResource
   * @return a resource representing the given file
   */
  createMediaForExistingFilename: function (filename) {
    var file = this.project.projectFolder;
    file.append(filename);
    if (! file.exists())
      throw "File " + filename + " does not exist";

    // Ensure it's a supported media type
    var supported = { image: 1, audio: 1, video: 1 };
    var msvc = getMIMEService();
    var mimeType = msvc.getTypeFromFile(file);
    var type = mimeType.split('/').shift();
    if (! supported[type])
      throw gApp.getText("UnsupportedMediaMsg");

    var typeres;
    switch (type) {
      case "image": typeres = this.rdf.GetResource(Cx.NS_CX + "Image"); break;
      case "audio": typeres = this.rdf.GetResource(Cx.NS_CX + "Audio"); break;
      case "video": typeres = this.rdf.GetResource(Cx.NS_CX + "Video"); break;
    }

    var fileres = this.rdf.GetResource(this.project.mintURI());
    var typearc = this.rdf.GetResource(Cx.NS_RDF + "type");
    var titlearc = this.rdf.GetResource(Cx.NS_DC + "title");
    var filearc = this.rdf.GetResource(Cx.NS_CX + "localFile");

    setRDFString(this.ds, fileres, titlearc, file.leafName);
    setRDFString(this.ds, fileres, filearc, file.leafName);
    this.ds.Assert(fileres, typearc, typeres, true);

    return fileres;
  },


  /**
   * Add a media file to the project and create the corresponding resource.
   * @param file  the media file to add
   * @type nsIRDFResource
   * @return a resource representing the given file
   */
  addMediaFromFile: function (file) {
    var msvc = getMIMEService();
    var mimeType = msvc.getTypeFromFile(file);
    dump("--- addMediaFromFile: " + mimeType + "\n");
    if (mimeType == "application/octet-stream") {
      var ext = file.leafName.match(/\.([^\.]+)$/);
      if (ext) {
        switch (ext[1].toLowerCase()) {
          case "bmp":
            mimeType = "image/bmp"; break;
          case "jpg":
          case "jpeg":
            mimeType = "image/jpeg"; break;
          case "png":
            mimeType = "image/png"; break;
          case "gif":
            mimeType = "image/gif"; break;
        }
      }
    }
    var type = mimeType.split('/').shift();

    // Ensure it's a supported media type
    var supported = { image: 1, audio: 1, video: 1 };

    if (! supported[type])
      throw gApp.getText("UnsupportedMediaMsg");

    var typeres;
    switch (type) {
      case "image": typeres = this.rdf.GetResource(Cx.NS_CX + "Image"); break;
      case "audio": typeres = this.rdf.GetResource(Cx.NS_CX + "Audio"); break;
      case "video": typeres = this.rdf.GetResource(Cx.NS_CX + "Video"); break;
    }

    var dstfile = copyToUnique(file, this.project.projectFolder, file.leafName);

    var fileres = this.rdf.GetResource(this.project.mintURI());
    var typearc = this.rdf.GetResource(Cx.NS_RDF + "type");
    var titlearc = this.rdf.GetResource(Cx.NS_DC + "title");
    var filearc = this.rdf.GetResource(Cx.NS_CX + "localFile");

    setRDFString(this.ds, fileres, titlearc, file.leafName);
    setRDFString(this.ds, fileres, filearc, dstfile.leafName);
    this.ds.Assert(fileres, typearc, typeres, true);

    return fileres;
  },


  /**
   * Presents an open file dialog for the user to pick media files.
   * @param type  "all" (default), "image", "video", or "audio" (optional)
   * @param multiple  true if the user should be able to pick multiple files
   * @type nsIFile[]
   * @return A single nsIFile if multiple is false, an array of nsIFile if
   *         multiple is true, or null if the user cancelled.
   */
  showMediaPicker: function (type, multiple) {
    type = type ? type : "all";

    var fp = getFilePicker();
    var mode = multiple ? fp.modeOpenMultiple : fp.modeOpen;
    fp.init(window, gApp.getText("AddMedia"), mode);
    fp.displayDirectory = getMediaDir();
    if (type == "image")
      fp.appendFilters(fp.filterImages);
    else
      fp.appendFilters(fp.filterAll);

    if (fp.show() != fp.returnOK)
      return null;

    var supportedTypes = {
      image: (type == "image" || type == "all") ? true : false,
      audio: (type == "audio" || type == "all") ? true : false,
      video: (type == "video" || type == "all") ? true : false
    };

    var msvc = getMIMEService();
    var files = [];
    if (multiple) {
      var fileiter = fp.files;
      var ILocalFile = Components.interfaces.nsILocalFile;
      while (fileiter.hasMoreElements()) {
        var file = fileiter.getNext().QueryInterface(ILocalFile);
        var mimetype = msvc.getTypeFromFile(file);
        if (mimetype == "application/octet-stream") {
          var ext = file.leafName.match(/\.([^\.]+)$/);
          if (ext) {
            switch (ext[1].toLowerCase()) {
              case "bmp":
                mimetype = "image/bmp"; break;
              case "jpg":
              case "jpeg":
                mimetype = "image/jpeg"; break;
              case "png":
                mimetype = "image/png"; break;
              case "gif":
                mimetype = "image/gif"; break;
            }
          }
        }
        dump("--- showMediaPicker: " + mimetype + "\n");
        var maintype = mimetype.split("/").shift();
        if (supportedTypes[maintype])
          files.push(file);
      }
    }
    else {
      var mimetype = msvc.getTypeFromFile(fp.file);
      var maintype = mimetype.split("/").shift();
      if (supportedTypes[maintype])
        files.push(fp.file);
    }

    if (files.length == 0)
      throw gApp.getText("UnsupportedMediaMsg");

    setMediaDir(files[0].parent);

    return multiple ? files : files[0];
  }
};
